function isDirectory(test) {
    return test !== undefined;
}
export class ModuleTree {
    root = {
        directories: {},
        files: {},
        isWildcard: false,
        name: "",
        rawName: "",
    };
    addModuleToDirectory(directory, segments, module) {
        if (directory === undefined) {
            return;
        }
        const [segment, ...remainingSegments] = segments;
        if (segment === undefined) {
            throw new Error("segments array is empty");
        }
        if (remainingSegments.length === 0) {
            directory.files[segment.toLowerCase()] = {
                isWildcard: segment.startsWith("{"),
                module,
                name: segment.replace(/^\{(?<name>.*)\}$/u, "$<name>"),
                rawName: segment,
            };
            return;
        }
        // eslint-disable-next-line @typescript-eslint/no-unnecessary-condition
        directory.directories[segment.toLowerCase()] ??= {
            directories: {},
            files: {},
            isWildcard: segment.startsWith("{"),
            name: segment.replace(/^\{(?<name>.*)\}$/u, "$<name>"),
            rawName: segment,
        };
        this.addModuleToDirectory(directory.directories[segment.toLocaleLowerCase()], remainingSegments, module);
    }
    add(url, module) {
        this.addModuleToDirectory(this.root, url.split("/").slice(1), module);
    }
    removeModuleFromDirectory(directory, segments) {
        if (!isDirectory(directory)) {
            return;
        }
        const [segment, ...remainingSegments] = segments;
        if (segment === undefined) {
            return;
        }
        if (remainingSegments.length === 0) {
            // eslint-disable-next-line @typescript-eslint/no-dynamic-delete
            delete directory.files[segment.toLowerCase()];
            return;
        }
        this.removeModuleFromDirectory(directory.directories[segment.toLowerCase()], remainingSegments);
    }
    remove(url) {
        const segments = url.split("/").slice(1);
        this.removeModuleFromDirectory(this.root, segments);
    }
    // eslint-disable-next-line max-params
    buildMatch(directory, segment, pathVariables, matchedPath) {
        const match = directory.files[segment.toLowerCase()] ??
            Object.values(directory.files).find((file) => file.isWildcard);
        if (match === undefined) {
            return undefined;
        }
        if (match.isWildcard) {
            return {
                ...match,
                matchedPath: `${matchedPath}/${match.rawName}`,
                pathVariables: {
                    ...pathVariables,
                    [match.name]: segment,
                },
            };
        }
        return {
            ...match,
            matchedPath: `${matchedPath}/${match.rawName}`,
            pathVariables,
        };
    }
    // eslint-disable-next-line max-statements, max-params
    matchWithinDirectory(directory, segments, pathVariables, matchedPath) {
        if (segments.length === 0) {
            return undefined;
        }
        const [segment, ...remainingSegments] = segments;
        if (segment === undefined) {
            throw new Error("segment cannot be undefined but TypeScript doesn't know that");
        }
        if (remainingSegments.length === 0 ||
            (remainingSegments.length === 1 && remainingSegments[0] === "")) {
            return this.buildMatch(directory, segment, pathVariables, matchedPath);
        }
        const exactMatch = directory.directories[segment.toLowerCase()];
        if (isDirectory(exactMatch)) {
            return this.matchWithinDirectory(exactMatch, remainingSegments, pathVariables, `${matchedPath}/${segment}`);
        }
        const wildcardDirectory = Object.values(directory.directories).find((subdirectory) => subdirectory.isWildcard);
        if (wildcardDirectory) {
            return this.matchWithinDirectory(wildcardDirectory, remainingSegments, {
                ...pathVariables,
                [wildcardDirectory.name]: segment,
            }, `${matchedPath}/${wildcardDirectory.rawName}`);
        }
        return undefined;
    }
    match(url) {
        return this.matchWithinDirectory(this.root, url.split("/").slice(1), {}, "");
    }
    get routes() {
        const routes = [];
        function traverse(directory, path) {
            Object.values(directory.directories).forEach((subdirectory) => {
                traverse(subdirectory, `${path}/${subdirectory.rawName}`);
            });
            Object.values(directory.files).forEach((file) => {
                const methods = Object.entries(file.module).map(([method, implementation]) => [method, String(implementation)]);
                routes.push({
                    methods: Object.fromEntries(methods),
                    path: `${path}/${file.rawName}`,
                });
            });
        }
        // eslint-disable-next-line unicorn/consistent-function-scoping
        function stripBrackets(string) {
            return string.replaceAll(/\{|\}/gu, "");
        }
        traverse(this.root, "");
        // eslint-disable-next-line etc/no-assign-mutated-array
        return routes.sort((first, second) => stripBrackets(first.path).localeCompare(stripBrackets(second.path)));
    }
}
